////////////////////////////////////////////////////////////////////////////////
//
//  ****************************************************************************
//  * Unit Name : FWDebugerCore
//  * Purpose   :  
//  * Author    :  (Rouse_) 
//  * Copyright :  Fangorn Wizards Lab 1998 - 2012.
//  * Version   : 1.0
//  * Home Page : http://rouse.drkb.ru
//  * Home Blog : http://alexander-bagel.blogspot.com/
//  ****************************************************************************
//

unit FWDebugerCore;

interface

uses
  Windows,
  Classes,
  SysUtils;

type

  EDebugCoreException = class(Exception);

  //      

  TResumeAction = (
    raRun,              //  
    raTraceInto,        //        
    raStepOver,         //        
    raRunUntilReturn,   //       
    raStop              //  
  );

  //   

  TCreateThreadEvent = procedure(Sender: TObject; ThreadIndex: Integer;
    Data: TCreateThreadDebugInfo) of object;
  TCreateProcessEvent = procedure(Sender: TObject; ThreadIndex: Integer;
    Data: TCreateProcessDebugInfo) of object;
  TExitThreadEvent = procedure(Sender: TObject; ThreadIndex: Integer;
    Data: TExitThreadDebugInfo) of object;
  TExitProcessEvent = procedure(Sender: TObject; ThreadIndex: Integer;
    Data: TExitProcessDebugInfo) of object;
  TLoadDllEvent = procedure(Sender: TObject; ThreadIndex: Integer;
    Data: TLoadDLLDebugInfo) of object;
  TUnLoadDllEvent = procedure(Sender: TObject; ThreadIndex: Integer;
    Data: TUnloadDLLDebugInfo) of object;
  TDebugStringEvent = procedure(Sender: TObject; ThreadIndex: Integer;
    Data: TOutputDebugStringInfo) of object;
  TRipEvent = procedure(Sender: TObject; ThreadIndex: Integer;
    Data: TRIPInfo) of object;

  //    

  TDefaultExceptionEvent = procedure(Sender: TObject; ThreadIndex: Integer;
    ExceptionRecord: Windows.TExceptionRecord) of object;
  TBreakPointEvent = procedure(Sender: TObject; ThreadIndex: Integer;
    ExceptionRecord: Windows.TExceptionRecord; BreakPointIndex: Integer;
    var ReleaseBreakpoint: Boolean) of object;

  //     (   ,
  //     )

  TExceptionCode = (ecUnknown, ecBreakpoint, ecSingleStep, ecCtrlC,
    ecNonContinuable, ecGuard);

  //      ( )

  TBreakpointType = (
    btBreakpoint,           // WriteProcessMemoryEx + 0xCC
    btMemoryBreakpoint      // VirtualProtectEx + PAGE_GUARD
  );

  //        

  TInt3Breakpoint = record
    Address: Pointer;
    ByteCode: Byte;
  end;

  TMemotyBreakPoint = record
    Address: Pointer;
    Size: DWORD;
    BreakOnWrite: Boolean;
    RegionStart: Pointer;
    RegionSize: DWORD;
    PreviosRegionProtect: DWORD;
  end;

  TBreakpoint = packed record
    bpType: TBreakpointType;
    Description: ShortString;
    Active: Boolean;
    case Integer of
      0: (Int3: TInt3Breakpoint;);
      1: (Memory: TMemotyBreakPoint);
  end;

  TBreakpointList = array of TBreakpoint;

  //      .
  //  ,     DebugProcessData
  //  !!!
  //        ,
  //         CloseHandle()

  TDebugProcessData = record
    ProcessID,
    EntryPoint: DWORD;
    CreatedProcessHandle,
    CreatedThreadHandle,
    AttachedProcessHandle,
    AttachedThreadHandle,
    AttachedFileHandle: THandle;
  end;

  //       ,
  //        .
  //  ,     GetThreadData
  //  !!!
  //        ,
  //         CloseHandle()

  THWBPIndex = 0..3;
  THWBPSize = (hsByte, hdWord, hsDWord);
  THWBPMode = (hmExecute, hmWrite, hmIO, hmReadWrite);

  THardwareBreakpoint = packed record
    Address: array [THWBPIndex] of Pointer;
    Size: array [THWBPIndex] of THWBPSize;
    Mode: array [THWBPIndex] of THWBPMode;
    Description: array [THWBPIndex] of ShortString;
    Active: array [THWBPIndex] of Boolean;
  end;

  TThreadData = record
    ThreadID: DWORD;
    ThreadHandle: THandle;
    Breakpoint: THardwareBreakpoint;
  end;
  TThreadList = array of TThreadData;

  THardwareBreakpointEvent = procedure(Sender: TObject; ThreadIndex: Integer;
    ExceptionRecord: Windows.TExceptionRecord; BreakPointIndex: THWBPIndex;
    var ReleaseBreakpoint: Boolean) of object;

  //  
const
  EFLAGS_CF = $001;
  EFLAGS_PF = $004;
  EFLAGS_AF = $010;
  EFLAGS_ZF = $040;
  EFLAGS_SF = $080;
  EFLAGS_TF = $100;
  EFLAGS_IF = $200;
  EFLAGS_DF = $400;
  EFLAGS_OF = $800;
  //   ...

type
  TFWDebugerCore = class
  private
    FSetEntryPointBreakPoint: Boolean;  //   ,      
    FProcessInfo: TDebugProcessData;    //     
    FMainLoopWaitPeriod: DWORD;         //    
    FBreakpointList: TBreakpointList;   //    
    FRestoreBPIndex,                    //    
    FRestoreMBPIndex,                   //    
    FRestoredHWBPIndex,                 //    
    FRestoredThread: Integer;
    FCloseDebugProcess: Boolean;        //          
    FThreadList: TThreadList;           //     
    FContinueStatus: DWORD;             //     ContinueDebugEvent
    FResumeAction: TResumeAction;       //       
    FRemoveCurrentBreakpoint: Boolean;  //    
    //  
    FMainLoopFailed: TNotifyEvent;
    FIdle: TNotifyEvent;
    FCreateThread: TCreateThreadEvent;
    FCreateProcess: TCreateProcessEvent;
    FExitThread: TExitThreadEvent;
    FExitProcess: TExitProcessEvent;
    FLoadDll: TLoadDllEvent;
    FUnLoadDll: TUnLoadDllEvent;
    FDebugString: TDebugStringEvent;
    FRip: TRipEvent;
    FExceptioEvents: array [TExceptionCode] of TDefaultExceptionEvent;
    FBreakPoint: TBreakPointEvent;
    FHardwareBreakpoint: THardwareBreakpointEvent;
    function GetExceptionEvent(const Index: TExceptionCode): TDefaultExceptionEvent;
    procedure SetExceptionEvent(const Index: TExceptionCode;
      const Value: TDefaultExceptionEvent);
    procedure SetCloseDebugProcess(const Value: Boolean);
  protected
    //       
    procedure AddThread(ThreadID, ThreadHandle: DWORD);
    procedure RemoveThread(Index: Integer);
    function GetThreadIndex(ThreadID: DWORD): Integer;
  protected
    //     
    procedure DoCreateThread(DebugEvent: TDebugEvent);
    procedure DoCreateProcess(DebugEvent: TDebugEvent);
    procedure DoExitThread(DebugEvent: TDebugEvent);
    procedure DoExitProcess(DebugEvent: TDebugEvent);
    procedure DoLoadDll(DebugEvent: TDebugEvent);
    procedure DoUnLoadDll(DebugEvent: TDebugEvent);
    procedure DoDebugString(DebugEvent: TDebugEvent);
    procedure DoRip(DebugEvent: TDebugEvent);
    procedure DoIdle;
    procedure DoMainLoopFailed;
    procedure DoResumeAction(ThreadIndex: Integer);
  protected
  //     
    procedure CallUnhandledExceptionEvents(ThreadIndex: Integer;
      Code: TExceptionCode; DebugEvent: TDebugEvent);
    procedure ProcessExceptionBreakPoint(ThreadIndex: Integer;
      DebugEvent: TDebugEvent);
    procedure ProcessExceptionSingleStep(ThreadIndex: Integer;
      DebugEvent: TDebugEvent);
    procedure ProcessExceptionGuardPage(ThreadIndex: Integer;
      DebugEvent: TDebugEvent);
    function ProcessHardwareBreakpoint(ThreadIndex: Integer;
      DebugEvent: TDebugEvent): Boolean;
  protected
    //    :

    //     
    function AddNewBreakPoint(Value: TBreakpoint): Boolean;
    //    
    procedure CheckBreakpointIndex(Value: Integer);
    //        Memory BP
    function CheckIsAddrInRealMemoryBPRegion(BreakPointIndex: Integer;
      AAddr: DWORD): Boolean;
    //   BP   
    function GetBPIndex(BreakPointAddr: DWORD): Integer;
    //   MBP   
    function GetMBPIndex(BreakPointAddr: DWORD; FromIndex: Integer = 0): Integer;
    //      BP
    function IsBreakpointPresent(Value: TBreakpoint): Boolean;
    //   Int3 
    procedure ToggleInt3Breakpoint(Index: Integer; Active: Boolean);
    //      
    procedure ToggleMemoryBreakpoint(Index: Integer; Active: Boolean);
    //     
    procedure UpdateHardwareBreakpoints(ThreadIndex: Integer);
  protected
    //  :

    //       
    procedure SetSingleStepMode(ThreadIndex: Integer;
      RestoreEIPAfterBP: Boolean);

  public
    constructor Create(AMainLoopWaitPeriod: DWORD = INFINITE);
    destructor Destroy; override;

    // / 
    function AttachToProcess(ProcessID: DWORD;
      SentEntryPointBreakPoint: Boolean): Boolean;
    function DebugNewProcess(const FilePath: string;
      SentEntryPointBreakPoint: Boolean): Boolean;
    function StopDebug: Boolean;
    procedure RunMainLoop;

    //   
    function ReadData(AddrPrt, ResultPtr: Pointer; DataSize: Integer): Boolean;
    function ReadStringA(AddrPrt: Pointer; DataSize: Integer): AnsiString;
    function ReadStringW(AddrPrt: Pointer; DataSize: Integer): WideString;
    function WriteData(AddrPrt, DataPtr: Pointer; DataSize: Integer): Boolean;
    function GetDllName(lpImageName, lpBaseOfDll: Pointer;
      var Unicode: Boolean): AnsiString;
    procedure SetFlag(ThreadIndex: Integer; Flag: DWORD; Value: Boolean);
    function GetFlagValue(ThreadIndex: Integer; Flag: DWORD): Boolean;

    //    
    function SetBreakpoint(Address: DWORD; const Description: string): Boolean;
    function SetMemoryBreakpoint(Address: Pointer; Size: DWORD;
      BreakOnWrite: Boolean; const Description: string): Boolean;
    procedure RemoveBreakpoint(Index: Integer);
    procedure ToggleBreakpoint(Index: Integer; Active: Boolean);
    function BreakpointCount: Integer;
    function BreakpointItem(Index: Integer): TBreakpoint;

    procedure RemoveCurrentBreakpoint;

    //     
    procedure SetHardwareBreakpoint(ThreadIndex: Integer; Address: Pointer;
      Size: THWBPSize; Mode: THWBPMode; HWIndex: THWBPIndex;
      const Description: string);
    procedure ToggleHardwareBreakpoint(ThreadIndex: Integer; Index: THWBPIndex;
      Active: Boolean);
    procedure DropHardwareBreakpoint(ThreadIndex: Integer; Index: THWBPIndex);
    procedure DropAllHardwareBreakpoint(ThreadIndex: Integer);

    //   
    property OnMainLoopFailed: TNotifyEvent read FMainLoopFailed
      write FMainLoopFailed;
    property OnIdle: TNotifyEvent read FIdle write FIdle;

    //   
    property OnCreateThread: TCreateThreadEvent read FCreateThread write FCreateThread;
    property OnCreateProcess: TCreateProcessEvent read FCreateProcess write FCreateProcess;
    property OnExitThread: TExitThreadEvent read FExitThread write FExitThread;
    property OnExitProcess: TExitProcessEvent read FExitProcess write FExitProcess;
    property OnLoadDll: TLoadDllEvent read FLoadDll write FLoadDll;
    property OnUnloadDll: TUnLoadDllEvent read FUnLoadDll write FUnLoadDll;
    property OnDebugString: TDebugStringEvent read FDebugString write FDebugString;
    property OnRip: TRipEvent read FRip write FRip;

    //  
    property OnBreakPoint: TBreakPointEvent read FBreakPoint write FBreakPoint;
    property OnHardwareBreakpoint: THardwareBreakpointEvent
      read FHardwareBreakpoint write FHardwareBreakpoint;
    property OnUnknownException: TDefaultExceptionEvent index ecUnknown
      read GetExceptionEvent write SetExceptionEvent;
    property OnUnknownBreakPoint: TDefaultExceptionEvent index ecBreakpoint
      read GetExceptionEvent write SetExceptionEvent;
    property OnSingleStep: TDefaultExceptionEvent index ecSingleStep
      read GetExceptionEvent write SetExceptionEvent;
    property OnCtrlC: TDefaultExceptionEvent index ecCtrlC
      read GetExceptionEvent write SetExceptionEvent;
    property OnNonContinuable: TDefaultExceptionEvent index ecNonContinuable
      read GetExceptionEvent write SetExceptionEvent;
    property OnPageGuard: TDefaultExceptionEvent index ecGuard
      read GetExceptionEvent write SetExceptionEvent;

    //   
    function GetThreadData(ThreadIndex: Integer): TThreadData;
    property ContinueStatus: DWORD read FContinueStatus write FContinueStatus;
    property CloseDebugProcessOnFree: Boolean
      read FCloseDebugProcess write SetCloseDebugProcess;
    property DebugProcessData: TDebugProcessData read FProcessInfo;
    property MainLoopWaitPeriod: DWORD read FMainLoopWaitPeriod
      write FMainLoopWaitPeriod;
    property ResumeAction: TResumeAction read FResumeAction write FResumeAction;
  end;

  function Debuger: TFWDebugerCore;

implementation

uses
  RTLConsts, Math;

const
  BPOpcode: Byte = $CC;

var
  _Debuger: TFWDebugerCore = nil;

function Debuger: TFWDebugerCore;
begin
  if _Debuger = nil then
    _Debuger := TFWDebugerCore.Create(INFINITE);
  Result := _Debuger;
end;

function Check(Value: Boolean): Boolean;
begin
  {$WARN SYMBOL_PLATFORM OFF}
  Result := Win32Check(Value);
  {$WARN SYMBOL_PLATFORM ON}
end;

function CodeDataToExceptionCode(Value: DWORD): TExceptionCode;
const
  EXCEPTION_UNKNOWN = 0;
  ExceptionCodeData: array [TExceptionCode] of DWORD = (
    EXCEPTION_UNKNOWN,
    EXCEPTION_BREAKPOINT,
    EXCEPTION_SINGLE_STEP,
    DBG_CONTROL_C,
    EXCEPTION_NONCONTINUABLE_EXCEPTION,
    EXCEPTION_GUARD_PAGE);
var
  I: TExceptionCode;
begin
  Result := ecUnknown;
  for I := Low(TExceptionCode) to High(TExceptionCode) do
    if Value = ExceptionCodeData[I] then
    begin
      Result := I;
      Break;
    end;
end;

{ TFWDebugerCore }

function TFWDebugerCore.AddNewBreakPoint(Value: TBreakpoint): Boolean;
var
  Len: Integer;
begin
  Result := not IsBreakpointPresent(Value);
  if Result then
  begin
    Value.Active := True;
    Len := BreakpointCount;
    SetLength(FBreakpointList, Len + 1);
    FBreakpointList[Len] := Value;
  end;
end;

procedure TFWDebugerCore.AddThread(ThreadID, ThreadHandle: DWORD);
var
  Len: Integer;
begin
  Len := Length(FThreadList);
  SetLength(FThreadList, Len + 1);
  FThreadList[Len].ThreadID := ThreadId;
  FThreadList[Len].ThreadHandle := ThreadHandle;
end;

function TFWDebugerCore.AttachToProcess(ProcessID: DWORD;
  SentEntryPointBreakPoint: Boolean): Boolean;
begin
  Result := False;
  if FProcessInfo.ProcessID <> 0 then Exit;
  FSetEntryPointBreakPoint := SentEntryPointBreakPoint;
  FProcessInfo.ProcessID := ProcessID;
  Result := DebugActiveProcess(ProcessID);
end;

function TFWDebugerCore.BreakpointCount: Integer;
begin
  Result := Length(FBreakpointList);
end;

function TFWDebugerCore.BreakpointItem(Index: Integer): TBreakpoint;
begin
  CheckBreakpointIndex(Index);
  Result := FBreakpointList[Index];
end;

procedure TFWDebugerCore.CallUnhandledExceptionEvents(ThreadIndex: Integer;
  Code: TExceptionCode; DebugEvent: TDebugEvent);
begin
  if Assigned(FExceptioEvents[Code]) then
    FExceptioEvents[Code](Self, ThreadIndex,
      DebugEvent.Exception.ExceptionRecord);
end;

procedure TFWDebugerCore.CheckBreakpointIndex(Value: Integer);
begin
  if (Value < 0) or (Value >= BreakpointCount) then
    raise EDebugCoreException.CreateFmt(SListIndexError, [Value]);
end;

function TFWDebugerCore.CheckIsAddrInRealMemoryBPRegion(
  BreakPointIndex: Integer; AAddr: DWORD): Boolean;
begin
  CheckBreakpointIndex(BreakPointIndex);
  Result := AAddr >= DWORD(FBreakpointList[BreakPointIndex].Memory.Address);
  if Result then
    Result := AAddr < DWORD(FBreakpointList[BreakPointIndex].Memory.Address) +
      FBreakpointList[BreakPointIndex].Memory.Size;
end;

constructor TFWDebugerCore.Create(AMainLoopWaitPeriod: DWORD);

  function SetDebugPriv: Boolean;
  var
    Token: THandle;
    tkp: TTokenPrivileges;
  begin
    Result := false;
    if OpenProcessToken(GetCurrentProcess,
      TOKEN_ADJUST_PRIVILEGES or TOKEN_QUERY, Token) then
    begin
      if LookupPrivilegeValue(nil, PChar('SeDebugPrivilege'),
        tkp.Privileges[0].Luid) then
      begin
        tkp.PrivilegeCount := 1;
        tkp.Privileges[0].Attributes := SE_PRIVILEGE_ENABLED;
        Result := AdjustTokenPrivileges(Token, False,
          tkp, 0, PTokenPrivileges(nil)^, PDWord(nil)^);
      end;
    end;
  end;

begin
  if not SetDebugPriv then
    RaiseLastOSError;
  FMainLoopWaitPeriod := AMainLoopWaitPeriod;
  FRestoreBPIndex := -1;
  FRestoreMBPIndex := -1;
  FRestoredHWBPIndex := -1;
  FRestoredThread := -1;
  FCloseDebugProcess := True;
end;

function TFWDebugerCore.DebugNewProcess(const FilePath: string;
  SentEntryPointBreakPoint: Boolean): Boolean;
var
  PI: TProcessInformation;
  SI: TStartupInfo;
begin
  Result := False;
  if FProcessInfo.ProcessID <> 0 then Exit;
  FSetEntryPointBreakPoint := SentEntryPointBreakPoint;
  ZeroMemory(@SI, SizeOf(TStartupInfo));
  SI.cb := SizeOf(TStartupInfo);
  Result := CreateProcess(PChar(FilePath),
    nil, nil, nil, False, DEBUG_PROCESS or DEBUG_ONLY_THIS_PROCESS,
    nil, nil, SI, PI);
  if Result then
  begin
    FProcessInfo.ProcessID := PI.dwProcessId;
    FProcessInfo.CreatedProcessHandle := PI.hProcess;
    FProcessInfo.CreatedThreadHandle := PI.hThread;
  end;
end;

destructor TFWDebugerCore.Destroy;
begin
  StopDebug;
  inherited;
end;

procedure TFWDebugerCore.DoCreateProcess(DebugEvent: TDebugEvent);
begin

  //    
  FProcessInfo.AttachedFileHandle := DebugEvent.CreateProcessInfo.hFile;
  FProcessInfo.AttachedProcessHandle := DebugEvent.CreateProcessInfo.hProcess;
  FProcessInfo.AttachedThreadHandle := DebugEvent.CreateProcessInfo.hThread;
  FProcessInfo.EntryPoint := DWORD(DebugEvent.CreateProcessInfo.lpStartAddress);

  AddThread(DebugEvent.dwThreadId, FProcessInfo.AttachedThreadHandle);

  //  BreakPoint    
  if FSetEntryPointBreakPoint then
    SetBreakpoint(FProcessInfo.EntryPoint, 'Process Entry Point Breakpoint');

  if Assigned(FCreateProcess) then
  begin
    FCreateProcess(Self, GetThreadIndex(DebugEvent.dwThreadId),
      DebugEvent.CreateProcessInfo);
    DoResumeAction(GetThreadIndex(DebugEvent.dwThreadId));
  end;
end;

procedure TFWDebugerCore.DoCreateThread(DebugEvent: TDebugEvent);
begin
  AddThread(DebugEvent.dwThreadId, DebugEvent.CreateThread.hThread);
  if Assigned(FCreateThread) then
  begin
    FCreateThread(Self, GetThreadIndex(DebugEvent.dwThreadId),
      DebugEvent.CreateThread);
    DoResumeAction(GetThreadIndex(DebugEvent.dwThreadId));
  end;
end;

procedure TFWDebugerCore.DoDebugString(DebugEvent: TDebugEvent);
begin
  if Assigned(FDebugString) then
  begin
    FDebugString(Self, GetThreadIndex(DebugEvent.dwThreadId),
      DebugEvent.DebugString);
    DoResumeAction(GetThreadIndex(DebugEvent.dwThreadId));
  end;
end;

procedure TFWDebugerCore.DoExitProcess(DebugEvent: TDebugEvent);
begin
  if Assigned(FExitProcess) then
    FExitProcess(Self, GetThreadIndex(DebugEvent.dwThreadId),
    DebugEvent.ExitProcess);
  if FProcessInfo.AttachedFileHandle <> 0 then
  begin
    CloseHandle(FProcessInfo.AttachedFileHandle);
    FProcessInfo.AttachedFileHandle := 0;
  end;
end;

procedure TFWDebugerCore.DoExitThread(DebugEvent: TDebugEvent);
var
  ThreadIndex: Integer;
begin
  ThreadIndex := GetThreadIndex(DebugEvent.dwThreadId);
  try
    if Assigned(FExitThread) then
      FExitThread(Self, ThreadIndex, DebugEvent.ExitThread);
    if ThreadIndex >= 0 then
    begin
      if FProcessInfo.AttachedThreadHandle =
        FThreadList[ThreadIndex].ThreadHandle then
        FProcessInfo.AttachedThreadHandle := 0;
    end;
  finally
    RemoveThread(ThreadIndex);
  end;
end;

procedure TFWDebugerCore.DoIdle;
begin
  if Assigned(FIdle) then
    FIdle(Self);
end;

procedure TFWDebugerCore.DoLoadDll(DebugEvent: TDebugEvent);
begin
  if Assigned(FLoadDll) then
  begin
    FLoadDll(Self, GetThreadIndex(DebugEvent.dwThreadId),
      DebugEvent.LoadDll);
    DoResumeAction(GetThreadIndex(DebugEvent.dwThreadId));
  end;
  CloseHandle(DebugEvent.LoadDll.hFile);
end;

procedure TFWDebugerCore.DoMainLoopFailed;
begin
  if Assigned(FMainLoopFailed) then
    FMainLoopFailed(Self);
end;

procedure TFWDebugerCore.DoResumeAction(ThreadIndex: Integer);
begin
  //     
  //        
  //      TraceIn  StepOver  ..
  //        

  case ResumeAction of
    raTraceInto:
      SetSingleStepMode(ThreadIndex, False);

    raStepOver: {     } ;

      {      , .. StepOver
            ,  :
        CALL, INT, LOOP, LOOPZ, LOOPNZ, REP, REPZ, REPNZ, CMPS, CMPSB, CMPSW,
        LODSB, LODSW, MOVS, MOVSB, MOVSW, SCAS, SCASB, SCASW, STOS, STOSB, STOSW
      }

    raRunUntilReturn: {       ,    };

    raStop: ContinueStatus := DBG_CONTROL_C;
  end;
end;

procedure TFWDebugerCore.DoRip(DebugEvent: TDebugEvent);
begin
  if Assigned(FRip) then
    FRip(Self, GetThreadIndex(DebugEvent.dwThreadId),
      DebugEvent.RipInfo);
end;

procedure TFWDebugerCore.DoUnLoadDll(DebugEvent: TDebugEvent);
begin
  if Assigned(FUnLoadDll) then
  begin
    FUnLoadDll(Self, GetThreadIndex(DebugEvent.dwThreadId),
      DebugEvent.UnloadDll);
    DoResumeAction(GetThreadIndex(DebugEvent.dwThreadId));
  end;
end;

procedure TFWDebugerCore.DropAllHardwareBreakpoint(ThreadIndex: Integer);
var
  I: THWBPIndex;
  NeedUpdate: Boolean;
begin
  if ThreadIndex < 0 then Exit;
  NeedUpdate := False;
  for I := 0 to 3 do
    if FThreadList[ThreadIndex].Breakpoint.Address[I] <> nil then
    begin
      NeedUpdate := True;
      FThreadList[ThreadIndex].Breakpoint.Address[I] := nil;
    end;
  if NeedUpdate then
    UpdateHardwareBreakpoints(ThreadIndex);
end;

procedure TFWDebugerCore.DropHardwareBreakpoint(ThreadIndex: Integer;
  Index: THWBPIndex);
begin
  if ThreadIndex < 0 then Exit;
  if FThreadList[ThreadIndex].Breakpoint.Address[Index] = nil then Exit;
  FThreadList[ThreadIndex].Breakpoint.Address[Index] := nil;
  UpdateHardwareBreakpoints(ThreadIndex);
end;

function TFWDebugerCore.GetBPIndex(BreakPointAddr: DWORD): Integer;
var
  I: Integer;
begin
  Result := -1;
  for I := 0 to BreakpointCount - 1 do
  begin
    if FBreakpointList[I].bpType <> btBreakpoint then Continue;
    if FBreakpointList[I].Int3.Address = Pointer(BreakPointAddr) then
    begin
      Result := I;
      Break;
    end;
  end;
end;

  function GetMappedFileNameA(hProcess: THandle; lpv: Pointer;
    lpFilename: LPSTR; nSize: DWORD): DWORD; stdcall; external 'psapi.dll';

function TFWDebugerCore.GetDllName(lpImageName, lpBaseOfDll: Pointer;
  var Unicode: Boolean): AnsiString;
var
  DllNameAddr: Pointer;
  MappedName: array [0..MAX_PATH - 1] of AnsiChar;
begin 
  if ReadData(lpImageName, @DllNameAddr, 4) then
    Result := ReadStringA(DllNameAddr, MAX_PATH shl 1);
  if Result = '' then
  begin
    if GetMappedFileNameA(FProcessInfo.AttachedProcessHandle,
      lpBaseOfDll, @MappedName[0], MAX_PATH) > 0 then
    begin
      Result := PAnsiChar(@MappedName[0]);
      Unicode := False;
    end;
  end;
end;

function TFWDebugerCore.GetExceptionEvent(
  const Index: TExceptionCode): TDefaultExceptionEvent;
begin
  Result := FExceptioEvents[Index];
end;

function TFWDebugerCore.GetFlagValue(ThreadIndex: Integer; Flag: DWORD): Boolean;
var
  Context: TContext;
begin
  ZeroMemory(@Context, SizeOf(Context));
  Context.ContextFlags := CONTEXT_FULL;
  Check(GetThreadContext(FThreadList[ThreadIndex].ThreadHandle, Context));
  Result := Context.EFlags and Flag = 1;
end;

function TFWDebugerCore.GetMBPIndex(BreakPointAddr: DWORD;
  FromIndex: Integer): Integer;

  function CheckStartAddr(Data: Pointer): Boolean;
  begin
    Result := DWORD(Data) <= BreakPointAddr;
  end;

  function CheckEndAddr(Data: Pointer; Size: DWORD): Boolean;
  begin
    Result := DWORD(Data) + Size > BreakPointAddr;
  end;

var
  I: Integer;
begin
  Result := -1;
  for I := FromIndex to BreakpointCount - 1 do
  begin
    if FBreakpointList[I].bpType <> btMemoryBreakpoint then Continue;
    if
      CheckStartAddr(FBreakpointList[I].Memory.RegionStart) and
      CheckEndAddr(FBreakpointList[I].Memory.RegionStart,
        FBreakpointList[I].Memory.RegionSize) then
    begin
      Result := I;
      Break;
    end;
  end;
end;

function TFWDebugerCore.GetThreadData(ThreadIndex: Integer): TThreadData;
begin
  ZeroMemory(@Result, SizeOf(TThreadData));
  if (ThreadIndex >= 0) and (ThreadIndex < Length(FThreadList)) then
    Result := FThreadList[ThreadIndex];
end;

function TFWDebugerCore.GetThreadIndex(ThreadID: DWORD): Integer;
var
  I: Integer;
begin
  Result := -1;
  for I := 0 to Length(FThreadList) do
    if FThreadList[I].ThreadID = ThreadID then
    begin
      Result := I;
      Break;
    end;
end;

function TFWDebugerCore.IsBreakpointPresent(Value: TBreakpoint): Boolean;
var
  I: Integer;
begin
  Result := False;
  for I := 0 to BreakpointCount - 1 do
    case FBreakpointList[I].bpType of
      btBreakpoint:
        if FBreakpointList[I].Int3.Address = Value.Int3.Address then
        begin
          Result := True;
          Break;
        end;
      btMemoryBreakpoint:
        if
          CheckIsAddrInRealMemoryBPRegion(I, DWORD(Value.Memory.Address)) or
          CheckIsAddrInRealMemoryBPRegion(I, DWORD(Value.Memory.Address) +
            Value.Memory.Size - 1) then
        begin
          Result := True;
          Break;
        end;
    end;
end;

procedure TFWDebugerCore.ProcessExceptionBreakPoint(ThreadIndex: Integer;
  DebugEvent: TDebugEvent);
var
  ReleaseBP: Boolean;
  BreakPointIndex: Integer;
begin
  ReleaseBP := False;
  FRemoveCurrentBreakpoint := False;
  BreakPointIndex := GetBPIndex(
    DWORD(DebugEvent.Exception.ExceptionRecord.ExceptionAddress));
  if BreakPointIndex >= 0 then
  begin
    if Assigned(FBreakPoint) then
      FBreakPoint(Self, ThreadIndex,
        DebugEvent.Exception.ExceptionRecord, BreakPointIndex,
        ReleaseBP)
    else
      CallUnhandledExceptionEvents(ThreadIndex, ecBreakpoint, DebugEvent);

    ToggleInt3Breakpoint(BreakPointIndex, False);
    SetSingleStepMode(ThreadIndex, True);
    if ReleaseBP or FRemoveCurrentBreakpoint then
      RemoveBreakpoint(BreakPointIndex)
    else
      FRestoreBPIndex := BreakPointIndex;
  end
  else
    CallUnhandledExceptionEvents(ThreadIndex, ecBreakpoint, DebugEvent);
end;

procedure TFWDebugerCore.ProcessExceptionGuardPage(ThreadIndex: Integer;
  DebugEvent: TDebugEvent);
var
  CurrentMBPIndex: Integer;

  function CheckWriteMode: Boolean;
  begin
    Result := not FBreakpointList[CurrentMBPIndex].Memory.BreakOnWrite;
    if not Result then
      Result := DebugEvent.Exception.ExceptionRecord.ExceptionInformation[0] = 1;
  end;

var
  MBPIndex: Integer;
  ReleaseMBP: Boolean;
  dwGuardedAddr: DWORD;
begin
  ReleaseMBP := False;
  FRemoveCurrentBreakpoint := False;
  dwGuardedAddr :=
    DebugEvent.Exception.ExceptionRecord.ExceptionInformation[1];
  MBPIndex := GetMBPIndex(dwGuardedAddr);
  if MBPIndex >= 0 then
  begin
    CurrentMBPIndex := MBPIndex;
    while not CheckIsAddrInRealMemoryBPRegion(CurrentMBPIndex, dwGuardedAddr) do
    begin
      CurrentMBPIndex := GetMBPIndex(dwGuardedAddr, CurrentMBPIndex + 1);
      if CurrentMBPIndex < 0 then Break;
    end;

    if CurrentMBPIndex >= 0 then
    begin
      MBPIndex := CurrentMBPIndex;
      if Assigned(FBreakPoint) and CheckWriteMode then
        FBreakPoint(Self, ThreadIndex,
          DebugEvent.Exception.ExceptionRecord, MBPIndex, ReleaseMBP)
      else
        CallUnhandledExceptionEvents(ThreadIndex, ecGuard, DebugEvent);
    end
    else
      CallUnhandledExceptionEvents(ThreadIndex, ecGuard, DebugEvent);

    FBreakpointList[MBPIndex].Active := False;
    SetSingleStepMode(ThreadIndex, False);
    if ReleaseMBP or FRemoveCurrentBreakpoint then
      RemoveBreakpoint(MBPIndex)
    else
      FRestoreMBPIndex := MBPIndex;
  end
  else
    CallUnhandledExceptionEvents(ThreadIndex, ecGuard, DebugEvent);
end;

procedure TFWDebugerCore.ProcessExceptionSingleStep(ThreadIndex: Integer;
  DebugEvent: TDebugEvent);
var
  Handled: Boolean;
begin
  //  HWBP
  Handled := ProcessHardwareBreakpoint(ThreadIndex, DebugEvent);

  //    - HWPB   HWBP
  if not Handled and (FRestoredThread >= 0) and
    (FRestoredHWBPIndex >= 0) then
  begin
    ToggleHardwareBreakpoint(FRestoredThread, FRestoredHWBPIndex, True);
    FRestoredThread := -1;
    FRestoredHWBPIndex := -1;
  end;

  //  
  if FRestoreBPIndex >= 0 then
  begin
    CheckBreakpointIndex(FRestoreBPIndex);
    if FBreakpointList[FRestoreBPIndex].bpType = btBreakpoint then
      ToggleInt3Breakpoint(FRestoreBPIndex, True);
    FRestoreBPIndex := -1;
  end;

  //  M
  if FRestoreMBPIndex >= 0 then
  begin
    CheckBreakpointIndex(FRestoreMBPIndex);
    if FBreakpointList[FRestoreMBPIndex].bpType = btMemoryBreakpoint then
      ToggleMemoryBreakpoint(FRestoreMBPIndex, True);
    FRestoreMBPIndex := -1;
  end;

  //        
  //    
  if ResumeAction <> raRun then
  begin
    CallUnhandledExceptionEvents(ThreadIndex, ecSingleStep, DebugEvent);

    //         
    DoResumeAction(ThreadIndex);
  end;
end;

function TFWDebugerCore.ProcessHardwareBreakpoint(ThreadIndex: Integer;
  DebugEvent: TDebugEvent): Boolean;
var
  Index: Integer;
  Context: TContext;
  ReleaseBP: Boolean;
begin
  FRemoveCurrentBreakpoint := False;
  ZeroMemory(@Context, SizeOf(TContext));
  Context.ContextFlags := CONTEXT_DEBUG_REGISTERS;
  Check(GetThreadContext(FThreadList[ThreadIndex].ThreadHandle, Context));
  Result := Context.Dr6 and $F <> 0;
  if not Result then Exit;

  Index := -1;
  if Context.Dr6 and 1 <> 0 then
    Index := 0;
  if Context.Dr6 and 2 <> 0 then
    Index := 1;
  if Context.Dr6 and 4 <> 0 then
    Index := 2;
  if Context.Dr6 and 8 <> 0 then
    Index := 3;
  if Index < 0 then
  begin
    Result := False;
    Exit;
  end;

  ReleaseBP := False;

  if Assigned(FHardwareBreakpoint) then
    FHardwareBreakpoint(Self, ThreadIndex, DebugEvent.Exception.ExceptionRecord,
      Index, ReleaseBP);

  ToggleHardwareBreakpoint(ThreadIndex, Index, False);
  SetSingleStepMode(ThreadIndex, False);

  if ReleaseBP or FRemoveCurrentBreakpoint then
    DropHardwareBreakpoint(ThreadIndex, Index)
  else
  begin

    //   HWBP    ,
    //  ..    
    //  ProcessExceptionSingleStep,   HWBP  
    //        HWBP
    if (FRestoredThread >= 0) and (FRestoredHWBPIndex >= 0) then
      ToggleHardwareBreakpoint(FRestoredThread, FRestoredHWBPIndex, True);

    FRestoredHWBPIndex := Index;
    FRestoredThread := ThreadIndex;
  end;
end;

function TFWDebugerCore.ReadData(AddrPrt, ResultPtr: Pointer; DataSize: Integer): Boolean;
var
  Dummy: DWORD;
begin
  Result := ReadProcessMemory(FProcessInfo.AttachedProcessHandle,
    AddrPrt, ResultPtr, DataSize, Dummy) and (Integer(Dummy) = DataSize);
end;

function TFWDebugerCore.ReadStringA(AddrPrt: Pointer;
  DataSize: Integer): AnsiString;
begin
  SetLength(Result, DataSize);
  if not ReadData(AddrPrt, @Result[1], DataSize) then
    Result := '';
end;

function TFWDebugerCore.ReadStringW(AddrPrt: Pointer;
  DataSize: Integer): WideString;
begin
  SetLength(Result, DataSize);
  if not ReadData(AddrPrt, @Result[1], DataSize) then
    Result := '';
end;

procedure TFWDebugerCore.RemoveBreakpoint(Index: Integer);
var
  Len: Integer;
begin
  ToggleBreakpoint(Index, False);
  Len := BreakpointCount;
  if Len = 1 then
    SetLength(FBreakpointList, 0)
  else
  begin
    FBreakpointList[Index] := FBreakpointList[Len - 1];
    SetLength(FBreakpointList, Len - 1);
  end;
end;

procedure TFWDebugerCore.RemoveCurrentBreakpoint;
begin
  FRemoveCurrentBreakpoint := True;
end;

procedure TFWDebugerCore.RemoveThread(Index: Integer);
var
  I, Len: Integer;
begin
  Len := Length(FThreadList);
  if (Index >= 0) and (Index < Len) then
  begin
    for I := Index to Len - 2 do
      FThreadList[I] := FThreadList[I + 1];
    SetLength(FThreadList, Len - 1);
  end;
end;

procedure TFWDebugerCore.RunMainLoop;
var
  DebugEvent: TDebugEvent;
  CallNextLoopIteration: Boolean;
  ThreadIndex: Integer;
begin
  CallNextLoopIteration := False;
  repeat

    ContinueStatus := DBG_CONTINUE;

    if not WaitForDebugEvent(DebugEvent, MainLoopWaitPeriod) then
    begin
      if GetLastError = ERROR_SEM_TIMEOUT then
      begin
        DoIdle;
        if FProcessInfo.ProcessID = 0 then Exit;
        CallNextLoopIteration := True;
        Continue;
      end
      else
      begin
        DoMainLoopFailed;
        Break;
      end;
    end;
    
    case DebugEvent.dwDebugEventCode of

      CREATE_THREAD_DEBUG_EVENT:
        DoCreateThread(DebugEvent);

      CREATE_PROCESS_DEBUG_EVENT:
        DoCreateProcess(DebugEvent);

      EXIT_THREAD_DEBUG_EVENT:
        DoExitThread(DebugEvent);

      EXIT_PROCESS_DEBUG_EVENT:
      begin
        DoExitProcess(DebugEvent);
        Break;
      end;

      LOAD_DLL_DEBUG_EVENT:
        DoLoadDll(DebugEvent); 

      UNLOAD_DLL_DEBUG_EVENT:
        DoUnLoadDll(DebugEvent);

      OUTPUT_DEBUG_STRING_EVENT:
        DoDebugString(DebugEvent);

      RIP_EVENT:
        DoRip(DebugEvent);

      EXCEPTION_DEBUG_EVENT:
      begin
        ThreadIndex := GetThreadIndex(DebugEvent.dwThreadId);
        case DebugEvent.Exception.ExceptionRecord.ExceptionCode of

          EXCEPTION_BREAKPOINT:
            ProcessExceptionBreakPoint(ThreadIndex, DebugEvent);

          EXCEPTION_SINGLE_STEP:
            ProcessExceptionSingleStep(ThreadIndex, DebugEvent);

          EXCEPTION_GUARD_PAGE:
            ProcessExceptionGuardPage(ThreadIndex, DebugEvent);

        else
          CallUnhandledExceptionEvents(ThreadIndex, CodeDataToExceptionCode(
            DebugEvent.Exception.ExceptionRecord.ExceptionCode), DebugEvent);
        end;
      end;

    end;

    CallNextLoopIteration := ContinueDebugEvent(DebugEvent.dwProcessId,
      DebugEvent.dwThreadId, ContinueStatus);

  until not CallNextLoopIteration;
end;

function TFWDebugerCore.SetBreakpoint(Address: DWORD;
  const Description: string): Boolean;
var
  Breakpoint: TBreakpoint;
  OldProtect: DWORD;
  Dummy: DWORD;
begin
  ZeroMemory(@Breakpoint, SizeOf(TBreakpoint));
  Breakpoint.bpType := btBreakpoint;
  Breakpoint.Int3.Address := Pointer(Address);
  Breakpoint.Description := ShortString(Description);
  Check(VirtualProtectEx(FProcessInfo.AttachedProcessHandle,
    Pointer(Address), 1, PAGE_READWRITE, OldProtect));
  try
    Check(ReadProcessMemory(FProcessInfo.AttachedProcessHandle,
      Pointer(Address), @Breakpoint.Int3.ByteCode, 1, Dummy));
    Check(WriteProcessMemory(FProcessInfo.AttachedProcessHandle,
      Pointer(Address), @BPOpcode, 1, Dummy));
  finally
    Check(VirtualProtectEx(FProcessInfo.AttachedProcessHandle,
      Pointer(Address), 1, OldProtect, OldProtect));
  end;
  Result := AddNewBreakPoint(Breakpoint);
end;

  function DebugSetProcessKillOnExit(KillOnExit: BOOL): BOOL; stdcall;
    external kernel32;

procedure TFWDebugerCore.SetCloseDebugProcess(const Value: Boolean);
begin
  FCloseDebugProcess := Value;
  DebugSetProcessKillOnExit(CloseDebugProcessOnFree);
end;

procedure TFWDebugerCore.SetExceptionEvent(const Index: TExceptionCode;
  const Value: TDefaultExceptionEvent);
begin
  FExceptioEvents[Index] := Value;
end;

procedure TFWDebugerCore.SetFlag(ThreadIndex: Integer; Flag: DWORD;
  Value: Boolean);
var
  Context: TContext;
begin
  ZeroMemory(@Context, SizeOf(Context));
  Context.ContextFlags := CONTEXT_FULL;
  Check(GetThreadContext(FThreadList[ThreadIndex].ThreadHandle, Context));
  if Value then
    Context.EFlags := Context.EFlags or Flag
  else
    Context.EFlags := Context.EFlags and not Flag;
  Check(SetThreadContext(FThreadList[ThreadIndex].ThreadHandle, Context));
end;

procedure TFWDebugerCore.SetHardwareBreakpoint(ThreadIndex: Integer;
  Address: Pointer; Size: THWBPSize; Mode: THWBPMode;
  HWIndex: THWBPIndex; const Description: string);
begin
  if ThreadIndex < 0 then Exit;
  FThreadList[ThreadIndex].Breakpoint.Address[HWIndex] := Address;
  FThreadList[ThreadIndex].Breakpoint.Size[HWIndex] := Size;
  FThreadList[ThreadIndex].Breakpoint.Mode[HWIndex] := Mode;
  FThreadList[ThreadIndex].Breakpoint.Description[HWIndex] :=
    ShortString(Description);
  FThreadList[ThreadIndex].Breakpoint.Active[HWIndex] := True;
  UpdateHardwareBreakpoints(ThreadIndex);
end;

function TFWDebugerCore.SetMemoryBreakpoint(Address: Pointer; Size: DWORD;
  BreakOnWrite: Boolean; const Description: string): Boolean;
var
  Breakpoint: TBreakpoint;
  MBI: TMemoryBasicInformation;
  Index: Integer;
begin
  Index := GetMBPIndex(DWORD(Address));
  if (Index >= 0) and (FBreakpointList[Index].bpType = btMemoryBreakpoint) then
  begin
    MBI.BaseAddress := FBreakpointList[Index].Memory.RegionStart;
    MBI.RegionSize := FBreakpointList[Index].Memory.RegionSize;
    MBI.Protect := FBreakpointList[Index].Memory.PreviosRegionProtect;
  end
  else
    Check(VirtualQueryEx(DebugProcessData.AttachedProcessHandle,
      Address, MBI, SizeOf(TMemoryBasicInformation)) > 0);
  ZeroMemory(@Breakpoint, SizeOf(TBreakpoint));
  Breakpoint.bpType := btMemoryBreakpoint;
  Breakpoint.Description := ShortString(Description);
  Breakpoint.Memory.Address := Address;
  Breakpoint.Memory.Size := Size;
  Breakpoint.Memory.BreakOnWrite := BreakOnWrite;
  Breakpoint.Memory.RegionStart := MBI.BaseAddress;
  Breakpoint.Memory.RegionSize := MBI.RegionSize;
  if Size = 0 then
    Inc(Size);
  Check(VirtualProtectEx(FProcessInfo.AttachedProcessHandle,
    Address, Size, MBI.Protect or PAGE_GUARD,
    Breakpoint.Memory.PreviosRegionProtect));
  if Index >= 0 then
    Breakpoint.Memory.PreviosRegionProtect := MBI.Protect;
  Result := AddNewBreakPoint(Breakpoint);
end;

  function DebugActiveProcessStop(dwProcessId: DWORD): BOOL; stdcall;
    external kernel32;

procedure TFWDebugerCore.SetSingleStepMode(ThreadIndex: Integer;
  RestoreEIPAfterBP: Boolean);
var
  Context: TContext;
begin
  ZeroMemory(@Context, SizeOf(TContext));
  Context.ContextFlags := CONTEXT_FULL;
  Check(GetThreadContext(FThreadList[ThreadIndex].ThreadHandle, Context));
  if RestoreEIPAfterBP then
    Dec(Context.Eip);
  Context.EFlags := Context.EFlags or EFLAGS_TF;
  Check(SetThreadContext(FThreadList[ThreadIndex].ThreadHandle, Context));
end;

function TFWDebugerCore.StopDebug: Boolean;
begin
  Result := True;

  if CloseDebugProcessOnFree then
    if FProcessInfo.CreatedProcessHandle <> 0 then
      TerminateProcess(FProcessInfo.CreatedProcessHandle, 0);

  if FProcessInfo.CreatedProcessHandle <> 0 then
    CloseHandle(FProcessInfo.CreatedProcessHandle);
  if FProcessInfo.CreatedThreadHandle <> 0 then
    CloseHandle(FProcessInfo.CreatedThreadHandle);
  if FProcessInfo.AttachedFileHandle <> 0 then
    CloseHandle(FProcessInfo.AttachedFileHandle);

  if FProcessInfo.ProcessID = 0 then Exit;
  Result := DebugActiveProcessStop(FProcessInfo.ProcessID);
  ZeroMemory(@FProcessInfo, SizeOf(TDebugProcessData));
end;

procedure TFWDebugerCore.ToggleBreakpoint(Index: Integer; Active: Boolean);
begin
  CheckBreakpointIndex(Index);
  case FBreakpointList[Index].bpType of
    btBreakpoint: ToggleInt3Breakpoint(Index, Active);
    btMemoryBreakpoint: ToggleMemoryBreakpoint(Index, Active);
  end;
end;

procedure TFWDebugerCore.ToggleHardwareBreakpoint(ThreadIndex: Integer;
  Index: THWBPIndex; Active: Boolean);
begin
  if ThreadIndex < 0 then Exit;
  if FThreadList[ThreadIndex].Breakpoint.Active[Index] = Active then Exit;
  FThreadList[ThreadIndex].Breakpoint.Active[Index] := Active;
  UpdateHardwareBreakpoints(ThreadIndex);
end;

procedure TFWDebugerCore.ToggleInt3Breakpoint(Index: Integer;
  Active: Boolean);
var
  OldProtect: DWORD;
  Dummy: DWORD;
begin
  CheckBreakpointIndex(Index);
  if FBreakpointList[Index].bpType <> btBreakpoint then Exit;
  if FBreakpointList[Index].Active = Active then Exit;
  Check(VirtualProtectEx(FProcessInfo.AttachedProcessHandle,
    FBreakpointList[Index].Int3.Address, 1, PAGE_READWRITE, OldProtect));
  try
    if Active then
      Check(WriteProcessMemory(FProcessInfo.AttachedProcessHandle,
        FBreakpointList[Index].Int3.Address, @BPOpcode, 1, Dummy))
    else
      Check(WriteProcessMemory(FProcessInfo.AttachedProcessHandle,
        FBreakpointList[Index].Int3.Address,
        @FBreakpointList[Index].Int3.ByteCode, 1, Dummy));
  finally
    Check(VirtualProtectEx(FProcessInfo.AttachedProcessHandle,
      FBreakpointList[Index].Int3.Address, 1, OldProtect, OldProtect));
  end;
  FBreakpointList[Index].Active := Active;
end;

procedure TFWDebugerCore.ToggleMemoryBreakpoint(Index: Integer;
  Active: Boolean);
var
  Dummy, TmpSize: DWORD;
begin
  CheckBreakpointIndex(Index);
  if FBreakpointList[Index].bpType <> btMemoryBreakpoint then Exit;
  if FBreakpointList[Index].Active = Active then Exit;
  TmpSize := FBreakpointList[Index].Memory.Size;
  if TmpSize = 0 then
    Inc(TmpSize);
  if Active then
    Check(VirtualProtectEx(FProcessInfo.AttachedProcessHandle,
      FBreakpointList[Index].Memory.Address,
      TmpSize,
      FBreakpointList[Index].Memory.PreviosRegionProtect or PAGE_GUARD, Dummy))
  else
    Check(VirtualProtectEx(FProcessInfo.AttachedProcessHandle,
      FBreakpointList[Index].Memory.Address,
      TmpSize,
      FBreakpointList[Index].Memory.PreviosRegionProtect, Dummy));
  FBreakpointList[Index].Active := Active;
end;

procedure TFWDebugerCore.UpdateHardwareBreakpoints(ThreadIndex: Integer);
const
  DR7_SET_LOC_DR0 = $01;
  DR7_SET_GLB_DR0 = $02;
  DR7_SET_LOC_DR1 = $04;
  DR7_SET_GLB_DR1 = $08;
  DR7_SET_LOC_DR2 = $10;
  DR7_SET_GLB_DR2 = $20;
  DR7_SET_LOC_DR3 = $40;
  DR7_SET_GLB_DR3 = $80;

  DR7_SET_LOC_ON  = $100;
  DR7_SET_GLB_ON  = $200;

  DR7_PROTECT     = $2000;

  DR_SIZE_BYTE    = 0;
  DR_SIZE_WORD    = 1;
  DR_SIZE_DWORD   = 3;

  DR_MODE_E       = 0;
  DR_MODE_W       = 1;
  DR_MODE_I       = 2;
  DR_MODE_R       = 3;

  DR7_MODE_DR0_E  = DR_MODE_E shl 16;
  DR7_MODE_DR0_W  = DR_MODE_W shl 16;
  DR7_MODE_DR0_I  = DR_MODE_I shl 16;
  DR7_MODE_DR0_R  = DR_MODE_R shl 16;

  DR7_SIZE_DR0_B  = DR_SIZE_BYTE shl 18;
  DR7_SIZE_DR0_W  = DR_SIZE_WORD shl 18;
  DR7_SIZE_DR0_D  = DR_SIZE_DWORD shl 18;

  DR7_MODE_DR1_E  = DR_MODE_E shl 20;
  DR7_MODE_DR1_W  = DR_MODE_W shl 20;
  DR7_MODE_DR1_I  = DR_MODE_I shl 20;
  DR7_MODE_DR1_R  = DR_MODE_R shl 20;

  DR7_SIZE_DR1_B  = DR_SIZE_BYTE shl 22;
  DR7_SIZE_DR1_W  = DR_SIZE_WORD shl 22;
  DR7_SIZE_DR1_D  = DR_SIZE_DWORD shl 22;

  DR7_MODE_DR2_E  = DR_MODE_E shl 24;
  DR7_MODE_DR2_W  = DR_MODE_W shl 24;
  DR7_MODE_DR2_I  = DR_MODE_I shl 24;
  DR7_MODE_DR2_R  = DR_MODE_R shl 24;

  DR7_SIZE_DR2_B  = DR_SIZE_BYTE shl 26;
  DR7_SIZE_DR2_W  = DR_SIZE_WORD shl 26;
  DR7_SIZE_DR2_D  = DR_SIZE_DWORD shl 26;

  DR7_MODE_DR3_E  = DR_MODE_E shl 28;
  DR7_MODE_DR3_W  = DR_MODE_W shl 28;
  DR7_MODE_DR3_I  = DR_MODE_I shl 28;
  DR7_MODE_DR3_R  = DR_MODE_R shl 28;

  DR7_SIZE_DR3_B  = DR_SIZE_BYTE shl 30;
  DR7_SIZE_DR3_W  = DR_SIZE_WORD shl 30;
  DR7_SIZE_DR3_D  = $C0000000; //DR_SIZE_DWORD shl 30;

  DR_On: array [THWBPIndex] of DWORD = (
    DR7_SET_LOC_DR0,
    DR7_SET_LOC_DR1,
    DR7_SET_LOC_DR2,
    DR7_SET_LOC_DR3
  );

  DR_Mode: array [THWBPIndex] of array [THWBPMode] of DWORD = (
    (DR7_MODE_DR0_E, DR7_MODE_DR0_W, DR7_MODE_DR0_I, DR7_MODE_DR0_R),
    (DR7_MODE_DR1_E, DR7_MODE_DR1_W, DR7_MODE_DR1_I, DR7_MODE_DR1_R),
    (DR7_MODE_DR2_E, DR7_MODE_DR2_W, DR7_MODE_DR2_I, DR7_MODE_DR2_R),
    (DR7_MODE_DR3_E, DR7_MODE_DR3_W, DR7_MODE_DR3_I, DR7_MODE_DR3_R)
  );

  DR_Size: array [THWBPIndex] of array [THWBPSize] of DWORD = (
    (DR7_SIZE_DR0_B, DR7_SIZE_DR0_W, DR7_SIZE_DR0_D),
    (DR7_SIZE_DR1_B, DR7_SIZE_DR1_W, DR7_SIZE_DR1_D),
    (DR7_SIZE_DR2_B, DR7_SIZE_DR2_W, DR7_SIZE_DR2_D),
    (DR7_SIZE_DR3_B, DR7_SIZE_DR3_W, DR7_SIZE_DR3_D)
  );

var
  Context: TContext;
  I: THWBPIndex;
begin
  if ThreadIndex < 0 then Exit;

  ZeroMemory(@Context, SizeOf(TContext));
  Context.ContextFlags := CONTEXT_DEBUG_REGISTERS;

  for I := 0 to 3 do
  begin
    if not FThreadList[ThreadIndex].Breakpoint.Active[I] then Continue;
    if FThreadList[ThreadIndex].Breakpoint.Address[I] <> nil then
    begin
      Context.Dr7 := Context.Dr7 or DR7_SET_LOC_ON;
      case I of
        0: Context.Dr0 := DWORD(FThreadList[ThreadIndex].Breakpoint.Address[I]);
        1: Context.Dr1 := DWORD(FThreadList[ThreadIndex].Breakpoint.Address[I]);
        2: Context.Dr2 := DWORD(FThreadList[ThreadIndex].Breakpoint.Address[I]);
        3: Context.Dr3 := DWORD(FThreadList[ThreadIndex].Breakpoint.Address[I]);
      end;
      Context.Dr7 := Context.Dr7 or DR_On[I];
      Context.Dr7 := Context.Dr7 or DR_Mode[I,
        FThreadList[ThreadIndex].Breakpoint.Mode[I]];
      Context.Dr7 := Context.Dr7 or DR_Size[I,
        FThreadList[ThreadIndex].Breakpoint.Size[I]];
    end;
  end;

  Check(SetThreadContext(FThreadList[ThreadIndex].ThreadHandle, Context));
end;

function TFWDebugerCore.WriteData(AddrPrt, DataPtr: Pointer;
  DataSize: Integer): Boolean;
var
  Dummy: DWORD;
begin
  Result := WriteProcessMemory(FProcessInfo.AttachedProcessHandle,
    AddrPrt, DataPtr, DataSize, Dummy) and (Integer(Dummy) = DataSize);
  if Result then
    Result := FlushInstructionCache(FProcessInfo.AttachedProcessHandle,
      AddrPrt, DataSize);
end;

initialization

  // Do nothing...

finalization

  _Debuger.Free;

end.
